//==================================================================
//
//	File name	:	sliderbar.js
//
//	Summary		:	Sliderbar class
//
//	Ceated		:	2018/12/05
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (sliderbar) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	sliderbar.GAUGE_TYPE_SHORT = 0;
	sliderbar.GAUGE_TYPE_LONG = 1;
	
	// Private variables
	let oSliding = null;
	let iSlidingPos = 0;
	let iLastGap = 0;
	
	//==========================================================
	//
	//	Class		:	SliderGauge
	//
	//	Summary		:	Class of slider gauge
	//
	//==========================================================
	class SliderGauge
	{
		//==================================================
		// constructor
		//==================================================
		constructor(iType,		// * Type of gauge[int]
					iPoint,		// * Point of gauge[int]
					iPos)		// * Position of gauge[int]
		{
			// Check parameters
			if(typeof iType != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_GAUGE,
											"sliderbar.js", "jaswin.sliderbar.SliderGauge.constructor");
			if((iType != sliderbar.GAUGE_TYPE_SHORT)
			&& (iType != sliderbar.GAUGE_TYPE_LONG))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_VALUE_GAUGE,
											"sliderbar.js", "jaswin.sliderbar.SliderGauge.constructor");
			
			if(typeof iPoint != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_POINT,
											"sliderbar.js", "jaswin.sliderbar.SliderGauge.constructor");
			
			if(typeof iPos != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT, jaswin.ALERT_TYPE_POSITION,
											"sliderbar.js", "jaswin.sliderbar.SliderGauge.constructor");
			
			// Set properties
			this.type = iType;
			this.point = iPoint;
			this.pos = iPos;
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	SliderBar
	//
	//	Summary		:	Class of SliderBar
	//
	//==========================================================
	sliderbar.SliderBar = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[FixedRect],
					iMinPoint,		// * Minimum point[int]
					iMaxPoint,		// * Maximum point[int]
					iStartPoint,	// * Start point[int]
					bEnabled,		// * Enabled or not[boolean]
					bGauge,			// * Show gauge or not[boolean]
					iShortGauge,	// Short gauge unit[int]
					iLongGauge,		// Long gauge unit[int]
					bBindGauge,		// Bind pointer to gauge or not[boolean]
					iTabIndex,		// Tab index[int]
					fpChange)		// Function on change
		{
			// Base class
			super(false, sID, false, oRect, bEnabled, iTabIndex);
			
			if(typeof iMinPoint != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MIN_POINT, sID),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			
			// Check parameters
			if(typeof iMaxPoint != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_MAX_POINT, sID),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			
			if(iMinPoint >= iMaxPoint)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_MIN_MAX, sID),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			
			if(typeof iStartPoint != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_START_POINT, sID),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			
			if(typeof bGauge != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SHOW_GAUGE, sID),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			
			if(bGauge)
			{
				if(typeof iShortGauge != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_SHORT_GAUGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
				if(iShortGauge < 0)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_SHORT_GAUGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
				
				if(typeof iLongGauge != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_LONG_GAUGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
				if(iLongGauge < 0)
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_LONG_GAUGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
				
				if((iShortGauge > 0)
				&& (iLongGauge > 0))
				{
					if(iLongGauge % iShortGauge)
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_VALUE_GAUGE_UNIT, sID),
													"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
				}
				
				if(typeof bBindGauge != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_BIND_GAUGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"sliderbar.js", "jaswin.sliderbar.SliderBar.constructor");
			}
			
			// Set properties
			this.minPoint = iMinPoint;		// Minimum point[int]
			this.maxPoint = iMaxPoint;		// Maximum point[int]
			this.point = iStartPoint;		// Start point[int]
			this.bindGauge = bBindGauge;	// Bind pointer to gauge or not[boolean]
			this.onChange = fpChange;		// Function on click[function(Argument:SliderBar class)]
			this.bar = null;				// Bar element[DOM Element]
			this.pointer = null;			// Pointer element[DOM Element]
			this.gauges = new Array();		// Array of gauge class
			
			// Create element
			this.element = document.createElement("div");
			this.element.style.position = "absolute";
			this.focusElement = this.element;
			
			if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
				this.element.style.right = this.rect.posX + "px";
			else
				this.element.style.left = this.rect.posX + "px";
			
			if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
				this.element.style.bottom = this.rect.posY + "px";
			else
				this.element.style.top = this.rect.posY + "px";
			
			this.element.style.width = this.rect.width - 2 + "px";		// 2:border
			
			// Create gauge
			if(bGauge)
			{
				this.element.style.height = "20px";
				
				let iCounter;
				let iCnt;
				let oGauge;
				let oSpan;
				let iType;
				let iPoint;
				let iPos;
				
				if((iShortGauge == 0) && (iLongGauge == 0))
					iCnt = 0;
				else if(iShortGauge == 0)
					iCnt = Math.floor((this.maxPoint - this.minPoint) / iLongGauge);
				else
					iCnt = Math.floor((this.maxPoint - this.minPoint) / iShortGauge);
				
				// First gauge
				oGauge = new SliderGauge(sliderbar.GAUGE_TYPE_LONG, this.minPoint, 0);
				oSpan = document.createElement("span");
				oSpan.className = "v-line-inset";
				oSpan.style.position = "absolute";
				oSpan.style.left = "4px";
				oSpan.style.top = "0px";
				oSpan.style.height = "8px";
				this.element.appendChild(oSpan);
				this.gauges[0] = oGauge;
				
				for(iCounter = 1; iCounter < iCnt; iCounter++)
				{
					if(iShortGauge != 0)
					{
						if(iCounter * iShortGauge + this.minPoint >= this.maxPoint)
							break;
					}
					else if(iLongGauge != 0)
					{
						if(iCounter * iLongGauge + this.minPoint >= this.maxPoint)
							break;
					}
					
					if(iLongGauge == 0)
						iType = sliderbar.GAUGE_TYPE_SHORT;
					else if(iShortGauge == 0)
						iType = sliderbar.GAUGE_TYPE_LONG;
					else if((iCounter * iShortGauge) % iLongGauge == 0)
						iType = sliderbar.GAUGE_TYPE_LONG;
					else
						iType = sliderbar.GAUGE_TYPE_SHORT;
					
					if(iShortGauge != 0)
						iPoint = this.minPoint + (iCounter * iShortGauge);
					else if(iLongGauge != 0)
						iPoint = this.minPoint + (iCounter * iLongGauge);
					
					iPos = Math.round((iPoint - this.minPoint) / (this.maxPoint - this.minPoint) * (this.rect.width - 2 - 9));
					
					oGauge = new SliderGauge(iType, iPoint, iPos);
					
					oSpan = document.createElement("span");
					oSpan.className = "v-line-inset";
					oSpan.style.position = "absolute";
					oSpan.style.left = oGauge.pos + 4 + "px";
					if(oGauge.type == sliderbar.GAUGE_TYPE_LONG)
					{
						oSpan.style.top = "0px";
						oSpan.style.height = "8px";
					}
					else
					{
						oSpan.style.top = "4px";
						oSpan.style.height = "4px";
					}
					this.element.appendChild(oSpan);
					this.gauges[iCounter] = oGauge;
				}
				
				// Last gauge
				oGauge = new SliderGauge(sliderbar.GAUGE_TYPE_LONG, this.maxPoint, this.rect.width - 2 - 9);
				oSpan = document.createElement("span");
				oSpan.className = "v-line-inset";
				oSpan.style.position = "absolute";
				oSpan.style.left = oGauge.pos + 4 + "px";
				oSpan.style.top = "0px";
				oSpan.style.height = "8px";
				this.element.appendChild(oSpan);
				this.gauges[this.gauges.length] = oGauge;
			}
			else
				this.element.style.height = "11px";
			
			// Create bar
			this.bar = document.createElement("div");
			this.bar.className = "sliderbar";
			this.bar.style.left = "3px";
			if(bGauge)
				this.bar.style.top = "13px";
			else
				this.bar.style.top = "4px";
			this.bar.style.width = this.rect.width - 2 - 8 + "px";
			this.element.appendChild(this.bar);
			
			// Create pointer
			this.pointer = document.createElement("span");
			if(this.enabled)
				this.pointer.className = "sliderbar-pointer";
			else
				this.pointer.className = "sliderbar-pointer-disabled";
			
			this.pointer.style.left = this.getPosition() + "px";
			if(bGauge)
				this.pointer.style.top = "9px";
			else
				this.pointer.style.top = "0px";
			this.element.appendChild(this.pointer);
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.bar.addEventListener("click", this.clickEvent = (e) => {
					this.clickBar(e);
				}, false);
				this.pointer.addEventListener("mousedown", this.mouseDownEvent = (e) => {
					oSliding = this;
					iLastGap = 0;
					iSlidingPos = e.clientX;
					
					global.element.addEventListener("mouseup", endSlidePointer, false);
					global.element.addEventListener("mouseout", outSlidePointer, false);
					global.element.addEventListener("mousemove", slidePointer, false);
				}, false);
				this.element.addEventListener("keydown", this.keyEvent = (e) => {
					this.keyDown(e);
				}, false);
			}
		}
		
		//==================================================
		//	Enable this sliderbar
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				this.pointer.className = "sliderbar-pointer";
				if(!this.clickEvent)
				{
					this.bar.addEventListener("click", this.clickEvent = (e) => {
						this.clickBar(e);
					}, false);
				}
				if(!this.mouseDownEvent)
				{
					this.pointer.addEventListener("mousedown", this.mouseDownEvent = (e) => {
						oSliding = this;
						iLastGap = 0;
						iSlidingPos = e.clientX;
						
						global.element.addEventListener("mouseup", endSlidePointer, false);
						global.element.addEventListener("mouseout", outSlidePointer, false);
						global.element.addEventListener("mousemove", slidePointer, false);
					}, false);
				}
				if(!this.keyEvent)
				{
					this.element.addEventListener("keydown", this.keyEvent = (e) => {
						this.keyDown(e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this sliderbar
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.pointer.className = "sliderbar-pointer-disabled";
			
			if(this.clickEvent)
			{
				this.bar.removeEventListener("click", this.clickEvent, false);
				this.clickEvent = null;
			}
			if(this.mouseDownEvent)
			{
				this.pointer.removeEventListener("mousedown", this.mouseDownEvent, false);
				this.mouseDownEvent = null;
			}
			if(this.keyEvent)
			{
				this.element.removeEventListener("keydown", this.keyEvent, false);
				this.keyEvent = null;
			}
		}
		
		//==================================================
		//	Get position of pointer by point
		//==================================================
		getPosition()
		{
			return Math.round((this.point - this.minPoint) / (this.maxPoint - this.minPoint) * (this.rect.width - 2 - 9));
		}
		
		//==================================================
		//	Slide pointer of this sliderbar
		//==================================================
		slide(iGap)
		{
			let iTmpPos;
			
			iTmpPos = parseInt(this.pointer.style.left) + iGap;
			if(iTmpPos < 0)
				iTmpPos = 0;
			else if(iTmpPos > this.rect.width - 2 - 9)
				iTmpPos = this.rect.width - 2 - 9;
			
			this.pointer.style.left = iTmpPos + "px";
			this.point = Math.round(iTmpPos / (this.rect.width - 2 - 9) * (this.maxPoint - this.minPoint)) + this.minPoint;
			
			if(this.onChange) this.onChange(this);
		}
		
		//==================================================
		//	Set point directly
		//==================================================
		setPoint(iValue)
		{
			if(typeof iValue != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"sliderbar.js", "jaswin.sliderbar.SliderBar.setPoint");
			
			if(iValue < this.minPoint)
				this.point = this.minPoint;
			else if(iValue > this.maxPoint)
				this.point = this.maxPoint;
			else
				this.point = iValue;
			
			this.pointer.style.left = this.getPosition() + "px";
			if(this.onChange) this.onChange(this);
		}
		
		//==================================================
		//	Click sliderbar
		//==================================================
		clickBar(e)
		{
			let oRect;
			let iClickPos;
			let iPointerPos;
			let bToLeft = true;
			let iCounter = 0;
			let iIndex = 0;
			
			oRect = this.bar.getBoundingClientRect();
			iClickPos = e.clientX - oRect.left - 3;
			iPointerPos = parseInt(this.pointer.style.left);
			if(iClickPos < iPointerPos)
				bToLeft = true;
			else if(iClickPos >= iPointerPos + 9)
				bToLeft = false;
			else
				return;
			
			if(this.gauges.length == 0)
			{
				if(bToLeft)
				{
					this.point--;
					if(this.point < this.minPoint)
						this.point = this.minPoint;
				}
				else
				{
					this.point++;
					if(this.point > this.maxPoint)
						this.point = this.maxPoint;
				}
			}
			else
			{
				if(bToLeft)
					iIndex = 0;
				else
					iIndex = this.gauges.length - 1;
				
				for(iCounter = 0; iCounter < this.gauges.length; iCounter++)
				{
					if(bToLeft)
					{
						if(this.gauges[iCounter].point < this.point)
							iIndex = iCounter;
						else
							break;
					}
					else
					{
						if(this.gauges[iCounter].point > this.point)
						{
							iIndex = iCounter;
							break;
						}
					}
				}
				this.point = this.gauges[iIndex].point;
			}
			this.pointer.style.left = this.getPosition() + "px";
			if(this.onChange) this.onChange(this);
		}
		
		//==================================================
		//	End of slidingbar
		//==================================================
		endSliding()
		{
			if((this.gauges.length > 0) && (!this.bindGauge))
				return;
			
			let iCounter = 0;
			let iIndex = -1;
			let iIndex2 = -1;
			let iMinGap = -1;
			let iTmp = 0;
			
			for(iCounter = 0; iCounter < this.gauges.length; iCounter++)
			{
				iTmp = Math.abs(this.point - this.gauges[iCounter].point);
				if((iMinGap == -1) || (iTmp < iMinGap))
				{
					iMinGap = iTmp;
					iIndex = iCounter;
				}
				else if(iTmp == iMinGap)
				{
					iIndex2 = iCounter;
				}
			}
			if(iIndex == -1)
				return;
			
			if(iIndex2 != -1)
			{
				if(iLastGap > 0)
					iIndex = iIndex2;
			}
			
			this.point = this.gauges[iIndex].point;
			this.pointer.style.left = this.getPosition() + "px";
			if(this.onChange) this.onChange(this);
		}
		
		//==================================================
		//	Keydown on slidingbar
		//==================================================
		keyDown(e)
		{
			if(e.keyCode == 37)
			{
				// Left
				if(this.point <= this.minPoint)
					return true;
				
				if(this.gauges.length == 0)
				{
					// No gauge
					this.point--;
					this.pointer.style.left = this.getPosition() + "px";
					if(this.onChange) this.onChange(this);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				else
				{
					if(e.ctrlKey)
					{
						// With Ctrl key
						let iCounter;
						for(iCounter = this.gauges.length - 1; iCounter >= 0; iCounter--)
						{
							if(this.point > this.gauges[iCounter].point)
							{
								this.point = this.gauges[iCounter].point;
								this.pointer.style.left = this.getPosition() + "px";
								if(this.onChange) this.onChange(this);
								e.preventDefault();
								e.stopPropagation();
								return false;
							}
						}
					}
					else
					{
						this.point--;
						this.pointer.style.left = this.getPosition() + "px";
						if(this.onChange) this.onChange(this);
						e.preventDefault();
						e.stopPropagation();
						return false;
					}
				}
			}
			else if(e.keyCode == 39)
			{
				// Right
				if(this.point >= this.maxPoint)
					return true;
				
				if(this.gauges.length == 0)
				{
					// No gauge
					this.point++;
					this.pointer.style.left = this.getPosition() + "px";
					if(this.onChange) this.onChange(this);
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				else
				{
					if(e.ctrlKey)
					{
						// With Ctrl key
						let iCounter;
						for(iCounter = 0; iCounter < this.gauges.length; iCounter++)
						{
							if(this.point < this.gauges[iCounter].point)
							{
								this.point = this.gauges[iCounter].point;
								this.pointer.style.left = this.getPosition() + "px";
								if(this.onChange) this.onChange(this);
								e.preventDefault();
								e.stopPropagation();
								return false;
							}
						}
					}
					else
					{
						this.point++;
						this.pointer.style.left = this.getPosition() + "px";
						if(this.onChange) this.onChange(this);
						e.preventDefault();
						e.stopPropagation();
						return false;
					}
				}
			}
			
			return true;
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	slidePointer
	//
	//	Summary		:	Event handler of sliding pointer
	//
	//==========================================================
	function slidePointer(e)
	{
		let oTarget = null;
		let iCurX;
		let iCurY;
		let oSliderBar = null;
		let oRect = null;
		let oSelection = window.getSelection();
		
		oSelection.removeAllRanges();
		
		if(!oSliding)
		{
			endSlidePointer(e);
			return;
		}
		
		iCurX = e.clientX;
		iCurY = e.clientY;
		
		oRect = oSliding.element.getBoundingClientRect();
		if((iCurX < oRect.left)
		|| (iCurX > oRect.right)
		|| (iCurY < oRect.top)
		|| (iCurY > oRect.bottom))
		{
			endSlidePointer(e);
			return;
		}
		
		iLastGap = iCurX - iSlidingPos;
		oSliding.slide(iLastGap);
		iSlidingPos = iCurX;
	}
	
	
	//==========================================================
	//
	//	Function	:	outSlidePointer
	//
	//	Summary		:	Event handler of mouse out slider bar
	//
	//==========================================================
	function outSlidePointer(e)
	{
		if(!oSliding)
			return;
		
		let oRect = oSliding.element.getBoundingClientRect();
		let iCurX;
		let iCurY;
		
		iCurX = e.clientX;
		iCurY = e.clientY;
		
		if((iCurX < oRect.left)
		|| (iCurX > oRect.right)
		|| (iCurY < oRect.top)
		|| (iCurY > oRect.bottom))
		{
			endSlidePointer(e);
		}
	}
	
	
	//==========================================================
	//
	//	Function	:	endSlidePointer
	//
	//	Summary		:	Event handler of end sliding pointer
	//
	//==========================================================
	function endSlidePointer(e)
	{
		global.element.removeEventListener("mouseup", endSlidePointer, false);
		global.element.removeEventListener("mouseout", outSlidePointer, false);
		global.element.removeEventListener("mousemove", slidePointer, false);
		
		if(oSliding)
			oSliding.endSliding();
		
		oSliding = null;
		iLastGap = 0;
	}
	
}(jaswin.sliderbar = jaswin.sliderbar || {}));
