//==================================================================
//
//	File name	:	textbox.js
//
//	Summary		:	TextBox class
//
//	Ceated		:	2018/11/29
//	Written by	:	Yuji Iihara
//	Copyright	:	Softinc.
//
//==================================================================
"use strict";

(function (textbox) {
	//==========================================================
	//	Global
	//==========================================================
	// Const
	textbox.STYLE_DEFAULT = 0;
	textbox.STYLE_FLAT = 1;
	textbox.STYLE_NOWRAP = 2;
	textbox.STYLE_TAB_INPUT = 4;
	textbox.STYLE_LINE_NO = 8;
	textbox.STYLE_HIGHLIGHTING = 16;
	textbox.TYPE_NONE = 0;
	textbox.TYPE_NUMBER_ALPHA = 1;
	textbox.TYPE_NUMBER = 2;
	textbox.TYPE_FORMATTED_NUMBER = 3;
	textbox.TYPE_PASSWORD = 4;
	textbox.IME_AUTO = 0;
	textbox.IME_ACTIVE = 1;
	textbox.IME_INACTIVE = 2;
	textbox.IME_DISABLED = 3;
	textbox.MARK_CHARACTERS = {58:':', 59:';', 64:'@',
								186:':', 187:';', 188:',', 189:'-', 190:'.', 191:'/', 192:'@',
								219:'[', 220:'\\', 221:']', 222:'^', 226:'\\'};
	textbox.SHIFTED_CHARACTERS = {49:'!', 50:'\"', 51:'#', 52:'$', 53:'%', 54:'&', 55:'\'', 56:'(', 57:')',
									58:'*', 59:'+', 64:'`',
									186:'*', 187:'+', 188:'<', 189:'=', 190:'>', 191:'?', 192:'`',
									219:'{', 220:'|', 221:'}', 222:'~', 226:'_'};
	
	// Private variables
	const WORKER_SRC = String.raw`self.addEventListener("message", (e) => {
		let sDoc = e.data;
		
		sDoc = sDoc.replace(/&/g, "&amp;");
		sDoc = sDoc.replace(/</g, "&lt;");
		sDoc = sDoc.replace(/>/g, "&gt;");
		
		#EXEC_HIGHLIGHT
		
		sDoc = sDoc.replace(/\t/g, "<span class=\"editbox-tab\">\t</span>");
		sDoc = sDoc.replace(/　/g, "<span class=\"editbox-widespace\">　</span>");
		sDoc = sDoc.replace(/\n/g, "<span class=\"editbox-enter\"> </span>\n");
		
		#EXEC_SEARCH
		
		self.postMessage(sDoc);
	}, false);
	#HIGHLIGHT_FUNC
	#SEARCH_FUNC`;
	
	const SEARCH_FUNC = String.raw`function searchFunc(sDoc, sWord, bRegExp, bIgnoreCase)
	{
		let sRet = sDoc;
		let oRegEx;
		let sSearchWord;
		
		if(bRegExp)
			sSearchWord = sWord;
		else
		{
			sSearchWord = sWord.replace(/</g, "&lt;");
			sSearchWord = sSearchWord.replace(/>/g, "&gt;");
			sSearchWord = escapeRegExp(sSearchWord);
		}
		if(bIgnoreCase)
			oRegEx = new RegExp(sSearchWord, "gi");
		else
			oRegEx = new RegExp(sSearchWord, "g");
		
		sRet = sRet.replace(oRegEx,
							function(){
								let iTagStart;
								let iTagEnd;
								iTagStart = arguments[arguments.length - 1].indexOf("<", arguments[arguments.length - 2] + arguments[0].length);
								iTagEnd = arguments[arguments.length - 1].indexOf(">", arguments[arguments.length - 2] + arguments[0].length);
								if((iTagEnd == -1) || (iTagStart < iTagEnd))
									return "<span class=\"editbox-search\">" + arguments[0] + "</span>";
								return arguments[0];
							});
		return sRet;
	}
	function escapeRegExp(#ESCAPE_ARGS)
	#ESCAPE_FUNC`;
	
	let oCurrentWorker = null;
	
	
	//==========================================================
	//
	//	Class		:	TextBox
	//
	//	Summary		:	Class of textbox
	//
	//==========================================================
	textbox.TextBox = class extends jaswin.Control
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[FixedRect]
					iStyle,			// * Textbox style[int]
					bEnabled,		// * Enabled or not[boolean]
					bReadOnly,		// Read only or not[boolean]
					sValue,			// Initial value[String]
					iMaxLength,		// Max length[int]
					sTooltip,		// Tool tip text[String]
					iTabIndex,		// Tab index[int]
					iType,			// Text box type[int]
					iDigit,			// Number digit[int]
					iImeMode,		// IME mode[int]
					fpChange,		// Function on change[function](Textbox)
					fpFocus,		// Function on focus[function](Textbox)
					fpBlur,			// Function on blur[function](Textbox)
					fpInput)		// Function on input[function](Textbox)
		{
			// Base class
			super(false, sID, false, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"textbox.js", "jaswin.textbox.TextBox.constructor");
			
			if((iStyle != textbox.STYLE_DEFAULT)
			&& (iStyle != textbox.STYLE_FLAT))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_VALUE_STYLE, sID),
											"textbox.js", "jaswin.textbox.TextBox.constructor");
			
			if((typeof bReadOnly != "undefined") && (bReadOnly != null))
			{
				if(typeof bReadOnly != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_READONLY, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			if(sValue)
			{
				if(typeof sValue != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(iMaxLength)
			{
				if(typeof iMaxLength != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_MAX_LENGTH, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(sTooltip)
			{
				if(typeof sTooltip != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(iType)
			{
				if(typeof iType != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TYPE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
				
				if((iType != textbox.TYPE_NONE)
				&& (iType != textbox.TYPE_NUMBER_ALPHA)
				&& (iType != textbox.TYPE_NUMBER)
				&& (iType != textbox.TYPE_FORMATTED_NUMBER)
				&& (iType != textbox.TYPE_PASSWORD))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_TYPE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(iDigit)
			{
				if(iType == textbox.TYPE_FORMATTED_NUMBER)
				{
					if(typeof iDigit != "number")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_DIGIT, sID),
													"textbox.js", "jaswin.textbox.TextBox.constructor");
					if(iDigit < 0)
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_VALUE_DIGIT, sID),
													"textbox.js", "jaswin.textbox.TextBox.constructor");
				}
			}
			
			if(iImeMode)
			{
				if(typeof iImeMode != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_IME_MODE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
				
				if((iImeMode != textbox.IME_AUTO)
				&& (iImeMode != textbox.IME_ACTIVE)
				&& (iImeMode != textbox.IME_INACTIVE)
				&& (iImeMode != textbox.IME_DISABLED))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_IME_MODE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(fpFocus)
			{
				if(typeof fpFocus != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_FOCUS, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(fpBlur)
			{
				if(typeof fpBlur != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_BLUR, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			if(fpInput)
			{
				if(typeof fpInput != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_INPUT, sID),
												"textbox.js", "jaswin.textbox.TextBox.constructor");
			}
			
			// Set properties
			this.style = iStyle;					// TextBox style[int]
			this.maxLength = iMaxLength;			// Max length[int]
			this.readOnly = bReadOnly;				// Read only or not[boolean]
			if(iType)								// Text box type[int]
				this.type = iType;
			else
				this.type = textbox.TYPE_NONE;
			if(iDigit)								// Number digit[int]
				this.digit = iDigit;
			else
				this.digit = 0;
			
			this.onChange = fpChange;				// Function on change[function](Textbox)
			this.onFocus = fpFocus;					// Function on focus[function](Textbox)
			this.onBlur = fpBlur;					// Function on blur[function](Textbox)
			this.onInput = fpInput;					// Function on input[function](Textbox)
			
			this.backgroundColor = "";				// Back ground color on normal[string]
			this.focusedBackColor = "";				// Back ground color on focus[string]
			this.unusableCharacters = null;			// Unusable characters[string]
			this.suggestionList = null;				// List box for suggestion[ListBox]
			
			// Create element
			this.element = document.createElement("input");
			if(this.type == textbox.TYPE_PASSWORD)
				this.element.type = "password";
			else
				this.element.type = "text";
			if(this.maxLength)
				this.element.maxLength = this.maxLength;
			
			if(this.style == textbox.STYLE_FLAT)
				this.element.className = "textbox-flat";
			else
				this.element.className = "textbox";
			
			if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
				this.element.style.right = this.rect.posX + "px";
			else
				this.element.style.left = this.rect.posX + "px";
			
			if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
				this.element.style.bottom = this.rect.posY + "px";
			else
				this.element.style.top = this.rect.posY + "px";
			
			if(this.style == textbox.STYLE_FLAT)
				this.element.style.width = this.rect.width - 2 + "px";				// 2:padding
			else
				this.element.style.width = this.rect.width - 2 - 2 + "px";			// 2:padding 2:border
			
			this.element.style.height = "16px";
			
			if(!this.enabled)
				this.element.disabled = true;
			
			if(this.readOnly)
				this.element.readOnly = true;
			
			if((this.type == textbox.TYPE_NUMBER_ALPHA)
			|| (this.type == textbox.TYPE_NUMBER)
			|| (this.type == textbox.TYPE_FORMATTED_NUMBER)
			|| (this.type == textbox.TYPE_PASSWORD))
			{
				this.element.style.imeMode = "disabled";
				
				if((this.type == textbox.TYPE_NUMBER)
				|| (this.type == textbox.TYPE_FORMATTED_NUMBER))
					this.element.style.textAlign = "right";
			}
			else
			{
				if(iImeMode == textbox.IME_ACTIVE)
					this.element.style.imeMode = "active";
				else if(iImeMode == textbox.IME_INACTIVE)
					this.element.style.imeMode = "inactive";
				else if(iImeMode == textbox.IME_DISABLED)
					this.element.style.imeMode = "disabled";
			}
			
			// Initial value
			if(sValue)
			{
				let fValue;
				
				if((this.type == textbox.TYPE_NUMBER)
				|| (this.type == textbox.TYPE_FORMATTED_NUMBER))
				{
					if(sValue.indexOf("."))
						fValue = parseFloat(sValue);
					else
						fValue = parseInt(sValue);
					
					if(!isNaN(fValue))
					{
						if(this.type == textbox.TYPE_NUMBER)
							this.element.value = "" + fValue;
						else
							this.element.value = jaswin.formatNumber("" + fValue, this.digit);
					}
				}
				else
					this.element.value = sValue;
			}
			
			this.focusElement = this.element;
			
			// Tooltip
			if((sTooltip) && (sTooltip.length > 0))
				this.element.placeholder = sTooltip;
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.focusElement.addEventListener("focus", this.focusEvent = (e) => {
					this.focus();
				}, false);
				this.focusElement.addEventListener("blur", this.blurEvent = (e) => {
					this.blur();
				}, false);
				this.element.addEventListener("keydown", this.keyDownEvent = (e) => {
					this.keyDown(e);
				}, false);
				this.element.addEventListener("change", this.changeEvent = (e) => {
					this.change();
				}, false);
				this.element.addEventListener("input", this.inputEvent = (e) => {
					this.input();
				}, false);
			}
		}
		
		//==================================================
		//	Enable this textbox
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(this.enabled)
			{
				this.element.disabled = false;
				
				if(!this.focusEvent)
				{
					this.focusElement.addEventListener("focus", this.focusEvent = (e) => {
						this.focus();
					}, false);
				}
				if(!this.blurEvent)
				{
					this.focusElement.addEventListener("blur", this.blurEvent = (e) => {
						this.blur();
					}, false);
				}
				if(!this.keyDownEvent)
				{
					this.element.addEventListener("keydown", this.keyDownEvent = (e) => {
						this.keyDown(e);
					}, false);
				}
				if(!this.changeEvent)
				{
					this.element.addEventListener("change", this.changeEvent = (e) => {
						this.change();
					}, false);
				}
				if(!this.inputEvent)
				{
					this.element.addEventListener("input", this.inputEvent = (e) => {
						this.input();
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this textbox
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.element.disabled = true;
			
			if(this.focusEvent)
			{
				this.focusElement.removeEventListener("focus", this.focusEvent, false);
				this.focusEvent = null;
			}
			if(this.blurEvent)
			{
				this.focusElement.removeEventListener("blur", this.blurEvent, false);
				this.blurEvent = null;
			}
			if(this.keyDownEvent)
			{
				this.element.removeEventListener("keydown", this.keyDownEvent, false);
				this.keyDownEvent = null;
			}
			if(this.changeEvent)
			{
				this.element.removeEventListener("change", this.changeEvent, false);
				this.changeEvent = null;
			}
			if(this.inputEvent)
			{
				this.element.removeEventListener("input", this.inputEvent, false);
				this.inputEvent = null;
			}
		}
		
		//==================================================
		//	Focus to this textbox
		//==================================================
		focus()
		{
			if(!this.enabled)
				return;
			
			if(this.style == textbox.STYLE_DEFAULT)
				this.element.className = "textbox-focus";
			
			this.element.style.backgroundColor = this.focusedBackColor;
			
			if(this.type == textbox.TYPE_FORMATTED_NUMBER)
				this.element.value = this.element.value.replace(/,/g, "");
			
			this.element.select();
			
			if(this.onFocus)
				this.onFocus(this);
		}
		
		//==================================================
		//	Blur from this textbox
		//==================================================
		blur()
		{
			if(!this.enabled)
				return;
			
			if((this.type == textbox.TYPE_FORMATTED_NUMBER)
			&& (this.element.value != ""))
			{
				let sValue = jaswin.formatNumber(this.element.value, this.digit);
				
				if(this.element.value != sValue)
					this.element.value = sValue;
			}
			
			if(this.style == textbox.STYLE_DEFAULT)
				this.element.className = "textbox";
			
			this.element.style.backgroundColor = this.backgroundColor;
			
			if(this.onBlur)
				this.onBlur(this);
		}
		
		//==================================================
		//	Input to this textbox
		//==================================================
		input()
		{
			if(this.onInput)
				this.onInput(this);
		}
		
		//==================================================
		//	Keydown this textbox
		//==================================================
		keyDown(e)
		{
			if((this.suggestionList)
			&& (this.suggestionList.element.style.display != "none")
			&& ((e.keyCode == 13) || (e.keyCode == 32) || (e.keyCode == 38) || (e.keyCode == 40)))
			{
				// When suggestion list is shown and key down enter/space or arrow up/down
				this.suggestionList.keyDownList(e);
			}
			
			if((this.suggestionList)
			&& (this.suggestionList.element.style.display != "none")
			&& (e.keyCode == 27))
			{
				// When suggestion list is shown and key down esc
				this.suggestionList.hide();
			}
			
			if((this.unusableCharacters)
			&& (this.unusableCharacters.length))
			{
				// Unusable characters
				let iCounter;
				let sUnusableChar;
				let sCurChar;
				
				for(iCounter = 0; iCounter < this.unusableCharacters.length; iCounter++)
				{
					sUnusableChar = this.unusableCharacters.charAt(iCounter);
					if((e.shiftKey) && !((e.keyCode >= 65) && (e.keyCode <= 90)))
					{
						// Shifted mark
						sCurChar = textbox.SHIFTED_CHARACTERS["" + e.keyCode];
						if((sCurChar) && (sCurChar == sUnusableChar))
						{
							e.preventDefault();
							e.stopPropagation();
							return false;
						}
					}
					else
					{
						if((e.keyCode >= 65) && (e.keyCode <= 90))
						{
							// [A-Z][a-z]
							sCurChar = String.fromCharCode(e.keyCode);
							if(sCurChar.toUpperCase() == sUnusableChar.toUpperCase())
							{
								e.preventDefault();
								e.stopPropagation();
								return false;
							}
						}
						else
						{
							// Others
							sCurChar = textbox.MARK_CHARACTERS["" + e.keyCode];
							if((sCurChar) && (sCurChar == sUnusableChar))
							{
								e.preventDefault();
								e.stopPropagation();
								return false;
							}
						}
					}
				}
			}
			
			if((this.type == textbox.TYPE_NUMBER)
			|| (this.type == textbox.TYPE_FORMATTED_NUMBER))
			{
				// Number only
				if(((e.keyCode < 48) || (e.keyCode > 57))
				&& ((e.keyCode < 96) || (e.keyCode > 105))
				&& ((e.keyCode < 33) || (e.keyCode > 46))
				&& (e.keyCode != 8)
				&& (e.keyCode != 9))
				{
					if(((e.keyCode == 190) && !e.shiftKey) || (e.keyCode == 110))
					{
						// Period
						if(this.element.value.indexOf(".") == -1)
							return true;
					}
					
					if(((e.keyCode == 189) && !e.shiftKey)
					|| ((e.keyCode == 173) && !e.shiftKey)
					|| (e.keyCode == 109))
					{
						// Minus
						if(((this.element.selectionEnd == 0) && (this.element.value.indexOf("-") == -1))
						|| ((this.element.selectionStart == 0) && (this.element.selectionEnd == this.element.value.length)))
							return true;
					}
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if((e.keyCode >= 48) && (e.keyCode <= 57) && e.shiftKey)
				{
					// Prevent Shift on number key
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if(((e.keyCode == 48) && !e.shiftKey) || (e.keyCode == 96))
				{
					// Zero
					if(this.element.value == "")
						return true;
					if((this.element.value.charAt(0) == "0")
					|| ((this.element.value.length >= 2) && (this.element.value.substring(0, 2) == "-0")))
					{
						if((this.element.value.indexOf(".") == -1)
						|| (this.element.selectionEnd <= this.element.value.indexOf(".")))
						{
							e.preventDefault();
							e.stopPropagation();
							return false;
						}
					}
				}
			}
			else if(this.type == textbox.TYPE_NUMBER_ALPHA)
			{
				// Number or Alphabet only
				if(((e.keyCode < 48) || (e.keyCode > 57))
				&& ((e.keyCode < 65) || (e.keyCode > 90))
				&& ((e.keyCode < 96) || (e.keyCode > 105))
				&& ((e.keyCode < 33) || (e.keyCode > 46))
				&& (e.keyCode != 8)
				&& (e.keyCode != 9))
				{
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
				
				if((e.keyCode >= 48) && (e.keyCode <= 57) && e.shiftKey)
				{
					// Prevent Shift on number key
					e.preventDefault();
					e.stopPropagation();
					return false;
				}
			}
			
			if(e.keyCode == 9)
			{
				if((this.suggestionList)
				&& (this.suggestionList.element.style.display != "none"))
					this.suggestionList.hide();
				
				window.getSelection().removeAllRanges();
			}
			
			return true;
		}
		
		//==================================================
		//	Change this textbox
		//==================================================
		change()
		{
			if(this.onChange)
				this.onChange(this);
		}
		
		//==================================================
		//	Get text
		//==================================================
		getText()
		{
			let fValue;
			
			if(this.element.value == "")
				return "";
			
			if(this.type == textbox.TYPE_FORMATTED_NUMBER)
			{
				if(this.element.value.indexOf("."))
					fValue = parseFloat(this.element.value);
				else
					fValue = parseInt(this.element.value);
				
				if(!isNaN(fValue))
					return jaswin.formatNumber("" + fValue, this.digit);
				else
					return "";
			}
			else
				return this.element.value;
		}
		
		//==================================================
		//	Get value
		//==================================================
		getValue()
		{
			if(this.type == textbox.TYPE_FORMATTED_NUMBER)
				return this.element.value.replace(/,/g, "");
			else
				return this.element.value;
		}
		
		//==================================================
		//	Set value
		//==================================================
		setValue(sValue)
		{
			if(typeof sValue != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setValue");
			
			if((this.type == textbox.TYPE_NUMBER)
			|| (this.type == textbox.TYPE_FORMATTED_NUMBER))
			{
				let fValue;
				
				if(sValue.indexOf("."))
					fValue = parseFloat(sValue);
				else
					fValue = parseInt(sValue);
				
				if(isNaN(fValue))
					this.element.value = "";
				else
				{
					if(this.type == textbox.TYPE_NUMBER)
						this.element.value = "" + fValue;
					else
						this.element.value = jaswin.formatNumber("" + fValue, this.digit);
				}
			}
			else
				this.element.value = sValue;
		}
		
		//==================================================
		//	Set this textbox read only or not
		//==================================================
		setReadOnly(bReadOnly)
		{
			if(typeof bReadOnly != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_READONLY, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setReadOnly");
			
			this.readOnly = bReadOnly;
			this.element.readOnly = bReadOnly;
		}
		
		//==================================================
		//	Set tooltip
		//==================================================
		setTooltip(sTooltip)
		{
			if(typeof sTooltip != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setTooltip");
			this.element.placeholder = sTooltip;
		}
		
		//==================================================
		//	Set font size
		//==================================================
		setFontSize(iFontPoint)
		{
			if(typeof iFontPoint != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_SIZE, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setFontSize");
			this.element.style.fontSize = iFontPoint + "pt";
		}
		
		//==================================================
		//	Set font color
		//==================================================
		setFontColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_COLOR),
											"textbox.js", "jaswin.textbox.TextBox.setFontColor");
			this.element.style.color = sColor;
		}
		
		//==================================================
		//	Set font family
		//==================================================
		setFontFamily(sFontFamily)
		{
			if(typeof sFontFamily != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_FAMILY, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setFontFamily");
			this.element.style.fontFamily = sFontFamily;
		}
		
		//==================================================
		//	Set background color
		//==================================================
		setBackgroundColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_BACKGROUND),
											"textbox.js", "jaswin.textbox.TextBox.setBackgroundColor");
			this.backgroundColor = sColor;
			this.element.style.backgroundColor = sColor;
		}
		
		//==================================================
		//	Set background color when focused
		//==================================================
		setFocusedBackgroundColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_BACKGROUND),
											"textbox.js", "jaswin.textbox.TextBox.setFocusedBackgroundColor");
			this.focusedBackColor = sColor;
		}
		
		//==================================================
		//	Set unusable characters
		//==================================================
		setUnusableCharacters(sUnusable)
		{
			if(typeof sUnusable != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_UNUSABLE_CHARACTERS),
											"textbox.js", "jaswin.textbox.TextBox.setUnusableCharacters");
			this.unusableCharacters = sUnusable;
		}
		
		//==================================================
		//	Set focus event
		//==================================================
		setFocusEvent(fpFocus)
		{
			if(typeof fpFocus != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_FOCUS, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setFocusEvent");
			this.onFocus = fpFocus;
		}
		
		//==================================================
		//	Set blur event
		//==================================================
		setBlurEvent(fpBlur)
		{
			if(typeof fpBlur != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_BLUR, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setBlurEvent");
			this.onBlur = fpBlur;
		}
		
		//==================================================
		//	Set change event
		//==================================================
		setChangeEvent(fpChange)
		{
			if(typeof fpChange != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setChangeEvent");
			
			this.onChange = fpChange;
		}
		
		//==================================================
		//	Set input event
		//==================================================
		setInputEvent(fpInput)
		{
			if(typeof fpInput != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_INPUT, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setInputEvent");
			this.onInput = fpInput;
		}
		
		//==================================================
		//	Set suggestion list
		//==================================================
		setSuggestionList(oListBox)
		{
			if(!jaswin.list)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "list.js"),
											"textbox.js", "jaswin.textbox.TextBox.setSuggestionList");
			
			if((typeof oListBox != "object")
			|| (!oListBox instanceof jaswin.list.ListBox))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LISTBOX, this.id),
											"textbox.js", "jaswin.textbox.TextBox.setSuggestionList");
			
			this.suggestionList = oListBox;
			this.suggestionList.hide();
		}
	}
	
	
	//==========================================================
	//
	//	Class		:	EditBox
	//
	//	Summary		:	Class of editbox
	//
	//==========================================================
	textbox.EditBox = class extends jaswin.Container
	{
		//==================================================
		// constructor
		//==================================================
		constructor(sID,			// * ID[String]
					oRect,			// * Size and position[ExpandableRect]
					iStyle,			// * Editbox style[int:bit flag]
					bEnabled,		// * Enabled or not[boolean]
					bReadOnly,		// Read only or not[boolean]
					sValue,			// Initial value[String]
					sTooltip,		// Tool tip text[String]
					iTabIndex,		// Tab index[int]
					iImeMode,		// IME mode[int]
					fpChange,		// Function on change[function](EditBox)
					fpFocus,		// Function on focus[function](EditBox)
					fpBlur,			// Function on blur[function](EditBox)
					fpInput,		// Function on input[function](EditBox)
					fpKeyDown,		// Function on keydown[function](EditBox, event)
					bDroppable,		// Droppable or not[boolean]
					fpDrop)			// Function on dropped[function]
		{
			// Base class
			super(sID, oRect, bEnabled, iTabIndex);
			
			// Check parameters
			if(typeof iStyle != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_STYLE, sID),
											"textbox.js", "jaswin.textbox.EditBox.constructor");
			
			if((typeof bReadOnly != "undefined") && (bReadOnly != null))
			{
				if(typeof bReadOnly != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_READONLY, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(sValue)
			{
				if(typeof sValue != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(sTooltip)
			{
				if(typeof sTooltip != "string")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(iImeMode)
			{
				if(typeof iImeMode != "number")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_IME_MODE, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
				
				if((iImeMode != textbox.IME_AUTO)
				&& (iImeMode != textbox.IME_ACTIVE)
				&& (iImeMode != textbox.IME_INACTIVE)
				&& (iImeMode != textbox.IME_DISABLED))
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_VALUE_IME_MODE, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(fpChange)
			{
				if(typeof fpChange != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(fpFocus)
			{
				if(typeof fpFocus != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_FOCUS, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(fpBlur)
			{
				if(typeof fpBlur != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_BLUR, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(fpInput)
			{
				if(typeof fpInput != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_INPUT, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(fpKeyDown)
			{
				if(typeof fpKeyDown != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_KEYDOWN, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
			}
			
			if(bDroppable)
			{
				if(typeof bDroppable != "boolean")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_DROPPABLE, sID),
												"textbox.js", "jaswin.textbox.EditBox.constructor");
				
				if(fpDrop)
				{
					if(typeof fpDrop != "function")
						throw new jaswin.JaswinError(jaswin.ERR_ALERT,
													jaswin.formatMessage(jaswin.ALERT_TYPE_ON_DROP, sID),
													"textbox.js", "jaswin.textbox.EditBox.constructor");
				}
			}
			
			// Set properties
			this.style = iStyle;			// EditBox style[int]
			this.readOnly = bReadOnly;		// Read only or not[boolean]
			this.onChange = fpChange;		// Function on change[function](EditBox)
			this.onFocus = fpFocus;			// Function on focus[function](EditBox)
			this.onBlur = fpBlur;			// Function on blur[function](EditBox)
			this.onInput = fpInput;			// Function on input[function](EditBox)
			this.onKeyDown = fpKeyDown;		// Function on keydown[function](EditBox, e)
			this.onHighlight = null;		// Function for highlighting[function](string) returns string
			this.textElement = null;		// Element of textarea[DOM Element]
			this.suggestionList = null;		// List box for suggestion[ListBox]
			this.searchWord = "";			// Search word[string]
			this.searchIgnoreCase = false;	// Serach word with ignore case option[boolean]
			this.searchRegExp = false;		// Serach word with regular expression option[boolean]
			this.droppable = false;			// Droppable or not[boolean]
			this.onDrop = null;				// Function on drop[function(Argument:EditBox class, e)]
			
			if(bDroppable)
			{
				this.droppable = bDroppable;
				this.onDrop = fpDrop;
			}
			
			// Create element
			this.element = document.createElement("div");
			if(this.style & textbox.STYLE_FLAT)
				this.element.className = "editbox-flat";
			else
				this.element.className = "editbox";
			
			// Horizontal rect
			if(this.rect.anchorX == jaswin.ANCHOR_BOTH)
			{
				this.element.style.left = this.rect.left + "px";
				this.element.style.right = this.rect.right + "px";
			}
			else
			{
				if(this.rect.anchorX == jaswin.ANCHOR_LEFT)
					this.element.style.left = this.rect.left + "px";
				else if(this.rect.anchorX == jaswin.ANCHOR_RIGHT)
					this.element.style.right = this.rect.right + "px";
				
				if(this.style & textbox.STYLE_FLAT)
					this.element.style.width = this.rect.width + "px";
				else
					this.element.style.width = this.rect.width - 2 + "px";
			}
			
			// Vertical rect
			if(this.rect.anchorY == jaswin.ANCHOR_BOTH)
			{
				this.element.style.top = this.rect.top + "px";
				this.element.style.bottom = this.rect.bottom + "px";
			}
			else
			{
				if(this.rect.anchorY == jaswin.ANCHOR_TOP)
					this.element.style.top = this.rect.top + "px";
				else if(this.rect.anchorY == jaswin.ANCHOR_BOTTOM)
					this.element.style.bottom = this.rect.bottom + "px";
				
				if(this.style & textbox.STYLE_FLAT)
					this.element.style.height = this.rect.height + "px";
				else
					this.element.style.height = this.rect.height - 2 + "px";
			}
			
			// Line no
			if(this.style & textbox.STYLE_LINE_NO)
			{
				this.lineNoElement = document.createElement("div");
				this.lineNoElement.className = "editbox-line-no";
				this.lineNoElement.innerHTML = "1";
				this.element.appendChild(this.lineNoElement);
			}
			
			let oDiv = document.createElement("div");
			oDiv.className = "editbox-inner";
			if((jaswin.BROWSER == "chrome")
			|| (jaswin.BROWSER == "opera"))
			{
				oDiv.style.right = "4px";
				oDiv.style.bottom = "4px";
			}
			if(this.style & textbox.STYLE_LINE_NO)
				oDiv.style.left = "46px";
			this.element.appendChild(oDiv);
			
			// View area(highlighting)
			if(this.style & textbox.STYLE_HIGHLIGHTING)
			{
				this.highlightElement = document.createElement("pre");
				this.highlightElement.className = "editbox-highlight";
				
				if(jaswin.BROWSER == "gecko")
					this.highlightElement.style.left = "1px";
				else if(jaswin.BROWSER == "edge")
				{
					this.highlightElement.style.left = "1px";
					this.highlightElement.style.top = "1px";
				}
				else if((jaswin.BROWSER == "chrome")
				|| (jaswin.BROWSER == "opera")
				|| (jaswin.BROWSER == "safari"))
				{
					this.highlightElement.style.left = "2px";
					this.highlightElement.style.top = "2px";
				}
				
				oDiv.appendChild(this.highlightElement);
			}
			
			// Input area
			this.textElement = document.createElement("textarea");
			this.textElement.className = "editbox-text";
			if(this.style & textbox.STYLE_HIGHLIGHTING)
				this.textElement.style.color = "transparent";
			
			if(this.style & textbox.STYLE_NOWRAP)
				this.textElement.wrap = "off";
			if(this.textElement.spellcheck)
				this.textElement.spellcheck = false;
			oDiv.appendChild(this.textElement);
			this.focusElement = this.textElement;
			
			if(!this.enabled)
				this.textElement.disabled = true;
			
			if(this.readOnly)
				this.textElement.readOnly = true;
			
			if(iImeMode == textbox.IME_ACTIVE)
				this.textElement.style.imeMode = "active";
			else if(iImeMode == textbox.IME_INACTIVE)
				this.textElement.style.imeMode = "inactive";
			else if(iImeMode == textbox.IME_DISABLED)
				this.textElement.style.imeMode = "disabled";
			
			if(sValue)
				this.textElement.value = sValue;
			
			// Tooltip
			if((sTooltip) && (sTooltip.length > 0))
				this.textElement.placeholder = sTooltip;
			
			// Tab index
			if(iTabIndex)
			{
				if(iTabIndex > 0)
				{
					this.tabIndex = iTabIndex;
					this.focusElement.tabIndex = this.tabIndex;
				}
			}
			
			// Event binding
			if(this.enabled)
			{
				this.focusElement.addEventListener("focus", this.focusEvent = (e) => {
					this.focus();
				}, false);
				this.focusElement.addEventListener("blur", this.blurEvent = (e) => {
					this.blur();
				}, false);
				this.textElement.addEventListener("keydown", this.keyDownEvent = (e) => {
					this.keyDown(e);
				}, false);
				this.textElement.addEventListener("change", this.changeEvent = (e) => {
					this.change();
				}, false);
				this.textElement.addEventListener("input", this.inputEvent = (e) => {
					this.input();
				}, false);
				if((this.style & textbox.STYLE_LINE_NO) || (this.style & textbox.STYLE_HIGHLIGHTING))
				{
					this.textElement.addEventListener("scroll", this.scrollEvent = (e) => {
						this.scroll();
					}, false);
				}
				this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
					let iCounter;
					let bFound = false;
					
					for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
					{
						if(e.dataTransfer.types[iCounter] == "Files")
							bFound = true;
					}
					if(!bFound)
						return;
					
					e.dataTransfer.dropEffect = "copy";
					e.stopPropagation();
					e.preventDefault();
				}, false);
				this.element.addEventListener("drop", this.dropEvent = (e) => {
					e.stopPropagation();
					e.preventDefault();
					if(this.onDrop)
						this.onDrop(this, e);
				}, false);
			}
		}
		
		//==================================================
		//	Enable this editbox
		//==================================================
		enable(bParent)
		{
			super.enable(bParent);
			
			if(bParent)
				this.parentEnabled = true;
			else
				this.enabled = true;
			
			if(this.enabled)
			{
				this.textElement.disabled = false;
				this.textElement.style.backgroundColor = "transparent";
				this.textElement.style.opacity = null;
				
				if(!this.focusEvent)
				{
					this.focusElement.addEventListener("focus", this.focusEvent = (e) => {
						this.focus();
					}, false);
				}
				if(!this.blurEvent)
				{
					this.focusElement.addEventListener("blur", this.blurEvent = (e) => {
						this.blur();
					}, false);
				}
				if(!this.keyDownEvent)
				{
					this.textElement.addEventListener("keydown", this.keyDownEvent = (e) => {
						this.keyDown(e);
					}, false);
				}
				if(!this.changeEvent)
				{
					this.textElement.addEventListener("change", this.changeEvent = (e) => {
						this.change();
					}, false);
				}
				if(!this.inputEvent)
				{
					this.textElement.addEventListener("input", this.inputEvent = (e) => {
						this.input();
					}, false);
				}
				if(((this.style & textbox.STYLE_LINE_NO) || (this.style & textbox.STYLE_HIGHLIGHTING))
				&& (!this.scrollEvent))
				{
					this.textElement.addEventListener("scroll", this.scrollEvent = (e) => {
						this.scroll();
					}, false);
				}
				if(!this.dragOverEvent)
				{
					this.element.addEventListener("dragover", this.dragOverEvent = (e) => {
						let iCounter;
						let bFound = false;
						
						for(iCounter = 0; iCounter < e.dataTransfer.types.length; iCounter++)
						{
							if(e.dataTransfer.types[iCounter] == "Files")
								bFound = true;
						}
						if(!bFound)
							return;
						
						e.dataTransfer.dropEffect = "copy";
						e.stopPropagation();
						e.preventDefault();
					}, false);
				}
				if(!this.dropEvent)
				{
					this.element.addEventListener("drop", this.dropEvent = (e) => {
						e.stopPropagation();
						e.preventDefault();
						if(this.onDrop)
							this.onDrop(this, e);
					}, false);
				}
			}
		}
		
		//==================================================
		//	Disable this editbox
		//==================================================
		disable(bParent)
		{
			super.disable(bParent);
			
			this.textElement.disabled = true;
			this.textElement.style.backgroundColor = "#c0c0c0";
			this.textElement.style.opacity = "0.4";
			
			if(this.focusEvent)
			{
				this.focusElement.removeEventListener("focus", this.focusEvent, false);
				this.focusEvent = null;
			}
			if(this.blurEvent)
			{
				this.focusElement.removeEventListener("blur", this.blurEvent, false);
				this.blurEvent = null;
			}
			if(this.keyDownEvent)
			{
				this.element.removeEventListener("keydown", this.keyDownEvent, false);
				this.keyDownEvent = null;
			}
			if(this.changeEvent)
			{
				this.element.removeEventListener("change", this.changeEvent, false);
				this.changeEvent = null;
			}
			if(this.inputEvent)
			{
				this.element.removeEventListener("input", this.inputEvent, false);
				this.inputEvent = null;
			}
			if(this.dragOverEvent)
			{
				this.element.removeEventListener("dragover", this.dragOverEvent, false);
				this.dragOverEvent = null;
			}
			if(this.dropEvent)
			{
				this.element.removeEventListener("drop", this.dropEvent, false);
				this.dropEvent = null;
			}
		}
		
		//==================================================
		//	Focus to this editbox
		//==================================================
		focus()
		{
			if(!this.enabled)
				return;
			
			if(!this.style & textbox.STYLE_FLAT)
				this.element.className = "editbox-focus";
			
			if(this.onFocus)
				this.onFocus(this);
		}
		
		//==================================================
		//	Blur from this editbox
		//==================================================
		blur()
		{
			if(!this.enabled)
				return;
			
			if(!this.style & textbox.STYLE_FLAT)
				this.element.className = "editbox";
			
			if(this.onBlur)
				this.onBlur(this);
		}
		
		//==================================================
		//	Input to this editbox
		//==================================================
		input()
		{
			if(this.style & textbox.STYLE_LINE_NO)
			{
				let iCnt;
				let iCounter;
				let sTmp = "";
				
				if(!this.textElement.value.match(/\n/g))
					iCnt = 1;
				else
					iCnt = this.textElement.value.match(/\n/g).length + 1;
				
				for(iCounter = 1; iCounter <= iCnt; iCounter++)
					sTmp += iCounter + "<br />";
				
				this.lineNoElement.innerHTML = sTmp;
			}
			
			if(this.style & textbox.STYLE_HIGHLIGHTING)
			{
				let sWork = WORKER_SRC;
				let sDoc = this.textElement.value;
				let sExecHighlight = "";
				let sHighlightFunc = "";
				let sArgs = "";
				let oFuncInfo;
				let sExecSearch = "";
				let sSearchFunc = "";
				let sUrl;
				let oWorker;
				let iIndex;
				
				this.textElement.style.color = null;
				
				if(this.onHighlight)
				{
					oFuncInfo = jaswin.parseFunction(this.onHighlight);
					sExecHighlight = "sDoc = highlightFunc(sDoc);";
					sArgs = oFuncInfo.args.join(",");
					sHighlightFunc = `function highlightFunc(${sArgs})
					${oFuncInfo.source}`;
				}
				sWork = sWork.replace("#EXEC_HIGHLIGHT", sExecHighlight);
				iIndex = sWork.indexOf("#HIGHLIGHT_FUNC");
				sWork = sWork.substring(0, iIndex) + sHighlightFunc + sWork.substring(iIndex + "#HIGHLIGHT_FUNC".length);
				
				if((this.searchWord) && (this.searchWord.length))
				{
					let sRegExpArg = "false";
					let sIgnoreCaseArg = "false";
					
					if(this.searchRegExp)
						sRegExpArg = "true";
					if(this.searchIgnoreCase)
						sIgnoreCaseArg = "true";
					sExecSearch = `sDoc = searchFunc(sDoc, "${this.searchWord}", ${sRegExpArg}, ${sIgnoreCaseArg});`;
					
					sSearchFunc = SEARCH_FUNC;
					oFuncInfo = jaswin.parseFunction(jaswin.escapeRegExp);
					sArgs = oFuncInfo.args.join(",");
					sSearchFunc = sSearchFunc.replace("#ESCAPE_ARGS", sArgs);
					sSearchFunc = sSearchFunc.replace("#ESCAPE_FUNC", oFuncInfo.source);
				}
				sWork = sWork.replace("#EXEC_SEARCH", sExecSearch);
				sWork = sWork.replace("#SEARCH_FUNC", sSearchFunc);
				
				sUrl = window.URL.createObjectURL(new Blob([sWork]), {type: "application/javascript; charset=utf-8"});
				oWorker = new Worker(sUrl);
				oCurrentWorker = oWorker;
				oWorker.addEventListener("message", (e) => {
					if(e.target == oCurrentWorker)
					{
						this.highlightElement.innerHTML = e.data;
						this.textElement.style.color = "transparent";
					}
				}, false);
				oWorker.postMessage(sDoc);
			}
			
			if(this.onInput)
				this.onInput(this);
		}
		
		//==================================================
		//	Scrolled this editbox
		//==================================================
		scroll()
		{
			if(this.style & textbox.STYLE_LINE_NO)
				this.lineNoElement.style.top = this.textElement.scrollTop * (-1) + "px";
			
			if(this.style & textbox.STYLE_HIGHLIGHTING)
			{
				if(jaswin.BROWSER == "gecko")
				{
					this.highlightElement.style.top = this.textElement.scrollTop * (-1) + "px";
					this.highlightElement.style.left = this.textElement.scrollLeft * (-1) + 1 + "px";
				}
				else if(jaswin.BROWSER == "edge")
				{
					this.highlightElement.style.top = this.textElement.scrollTop * (-1) + 1 + "px";
					this.highlightElement.style.left = this.textElement.scrollLeft * (-1) + 1 + "px";
				}
				else if((jaswin.BROWSER == "chrome")
				|| (jaswin.BROWSER == "opera")
				|| (jaswin.BROWSER == "safari"))
				{
					this.highlightElement.style.top = this.textElement.scrollTop * (-1) + 2 + "px";
					this.highlightElement.style.left = this.textElement.scrollLeft * (-1) + 2 + "px";
				}
				else
				{
					this.highlightElement.style.top = this.textElement.scrollTop * (-1) + "px";
					this.highlightElement.style.left = this.textElement.scrollLeft * (-1) + "px";
				}
			}
		}
		
		//==================================================
		//	Keydown this editbox
		//==================================================
		keyDown(e)
		{
			if((this.suggestionList)
			&& (this.suggestionList.element.style.display != "none")
			&& ((e.keyCode == 13) || (e.keyCode == 32) || (e.keyCode == 38) || (e.keyCode == 40)))
			{
				// When suggestion list is shown and key down enter/space or arrow up/down
				this.suggestionList.keyDownList(e);
			}
			
			if((this.suggestionList)
			&& (this.suggestionList.element.style.display != "none")
			&& (e.keyCode == 27))
			{
				// When suggestion list is shown and key down esc
				this.suggestionList.hide();
			}
			
			if(e.keyCode == 9)
			{
				if(this.style & textbox.STYLE_TAB_INPUT)
				{
					// Input a tab indent
					let iStart = this.textElement.selectionStart;
					let iEnd = this.textElement.selectionEnd;
					let sValue = this.textElement.value;
					let sSelected = this.textElement.value.substring(iStart, iEnd);
					let sBefore = this.textElement.value.substring(0, iStart);
					let sAfter = this.textElement.value.substring(iEnd);
					let iLineStart;
					let iLineEnd;
					let iCounter;
					
					if(iStart == iEnd)
					{
						this.textElement.value = "" + sBefore + "\t" + sAfter;
						this.textElement.selectionStart = this.textElement.selectionEnd = iStart + 1;
						e.preventDefault();
						e.stopPropagation();
						
						this.input();
						return false;
					}
					else
					{
						if(sSelected.indexOf("\n") == -1)
						{
							// selected one line
							iLineStart = sBefore.lastIndexOf("\n") + 1;
							iLineEnd = sValue.indexOf("\n", iEnd);
							if(iLineEnd == -1)
								iLineEnd = sValue.length;
							if(sSelected.length != iLineEnd - iLineStart)
							{
								this.textElement.value = "" + sBefore + "\t" + sAfter;
								this.textElement.selectionStart = this.textElement.selectionEnd = iStart + 1;
								e.preventDefault();
								e.stopPropagation();
								
								this.input();
								return false;
							}
						}
					}
					
					iStart = sBefore.lastIndexOf("\n") + 1;
					iEnd = sValue.indexOf("\n", iEnd);
					if(iEnd == -1)
						iEnd = sValue.length;
					sSelected = this.textElement.value.substring(iStart, iEnd);
					
					let saLines = sSelected.split("\n");
					iLineStart = iStart;
					for(iCounter = 0; iCounter < saLines.length; iCounter++)
					{
						this.textElement.value = this.textElement.value.substring(0, iLineStart)
												+ "\t" + this.textElement.value.substring(iLineStart);
						iLineStart = iLineStart + 1 + saLines[iCounter].length + 1;		// 1px:TAB 1px:LF
					}
					this.textElement.selectionStart = iStart;
					this.textElement.selectionEnd = iEnd + saLines.length;
					
					e.preventDefault();
					e.stopPropagation();
					
					this.input();
					return false;
				}
			}
			
			if(this.onKeyDown)
				return this.onKeyDown(this, e);
			
			return true;
		}
		
		//==================================================
		//	Change this editbox
		//==================================================
		change()
		{
			if(this.onChange)
				this.onChange(this);
		}
		
		//==================================================
		//	Set this editbox read only or not
		//==================================================
		setReadOnly(bReadOnly)
		{
			if(typeof bReadOnly != "boolean")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_READONLY, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setReadOnly");
			
			this.readOnly = bReadOnly;
			this.element.readOnly = bReadOnly;
		}
		
		//==================================================
		//	Get value
		//==================================================
		getValue()
		{
			return this.textElement.value;
		}
		
		//==================================================
		//	Set value
		//==================================================
		setValue(sValue)
		{
			if(typeof sValue != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_VALUE, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setValue");
			this.textElement.value = sValue;
			this.input();
			this.textElement.scrollTop = 0;
			this.textElement.selectionStart = 0;
			this.textElement.selectionEnd = 0;
		}
		
		//==================================================
		//	Replace selected string to upper case
		//==================================================
		toUpperCase()
		{
			if(this.textElement.selectionStart == this.textElement.selectionEnd)
				return;
			
			let iStart = this.textElement.selectionStart;
			let iEnd = this.textElement.selectionEnd;
			this.textElement.value = this.textElement.value.substring(0, iStart)
								+ this.textElement.value.substring(iStart, iEnd).toUpperCase()
								+ this.textElement.value.substring(iEnd);
			this.input();
			this.textElement.selectionStart = iStart;
			this.textElement.selectionEnd = iEnd;
		}
		
		//==================================================
		//	Replace selected string to lower case
		//==================================================
		toLowerCase()
		{
			if(this.textElement.selectionStart == this.textElement.selectionEnd)
				return;
			
			let iStart = this.textElement.selectionStart;
			let iEnd = this.textElement.selectionEnd;
			this.textElement.value = this.textElement.value.substring(0, iStart)
								+ this.textElement.value.substring(iStart, iEnd).toLowerCase()
								+ this.textElement.value.substring(iEnd);
			this.input();
			this.textElement.selectionStart = iStart;
			this.textElement.selectionEnd = iEnd;
		}
		
		//==================================================
		//	Set tooltip
		//==================================================
		setTooltip(sTooltip)
		{
			if(typeof sTooltip != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TOOLTIP, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setTooltip");
			this.textElement.placeholder = sTooltip;
		}
		
		//==================================================
		//	Set font size
		//==================================================
		setFontSize(iSize)
		{
			if(typeof iSize != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_SIZE, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setFontSize");
			this.textElement.style.fontSize = iSize + "pt";
			if(this.lineNoElement)
				this.lineNoElement.style.fontSize = iSize + "pt";
			if(this.highlightElement)
				this.highlightElement.style.fontSize = iSize + "pt";
		}
		
		//==================================================
		//	Set font color
		//==================================================
		setFontColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_COLOR),
											"textbox.js", "jaswin.textbox.EditBox.setFontColor");
			this.textElement.style.color = sColor;
		}
		
		//==================================================
		//	Set font family
		//==================================================
		setFontFamily(sFontFamily)
		{
			if(typeof sFontFamily != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_FONT_FAMILY, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setFontFamily");
			this.textElement.style.fontFamily = sFontFamily;
			if(this.lineNoElement)
				this.lineNoElement.style.fontFamily = sFontFamily;
			if(this.highlightElement)
				this.highlightElement.style.fontFamily = sFontFamily;
		}
		
		//==================================================
		//	Set tab indent size
		//==================================================
		setTabSize(iTabSize)
		{
			if(typeof iTabSize != "number")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_TAB_SIZE, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setFontSize");
			this.textElement.style.tabSize = "" + iTabSize;
			this.textElement.style.setProperty("-moz-tab-size", "" + iTabSize);
			if(this.highlightElement)
			{
				this.highlightElement.style.tabSize = "" + iTabSize;
				this.highlightElement.style.setProperty("-moz-tab-size", "" + iTabSize);
			}
		}
		
		//==================================================
		//	Set background color
		//==================================================
		setBackgroundColor(sColor)
		{
			if(typeof sColor != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_BACKGROUND),
											"textbox.js", "jaswin.textbox.EditBox.setBackgroundColor");
			this.backgroundColor = sColor;
			this.element.style.backgroundColor = sColor;
		}
		
		//==================================================
		//	Set search word
		//==================================================
		setSearchWord(sWord, bIgnoreCase, bRegExp)
		{
			if(!this.style & textbox.STYLE_HIGHLIGHTING)
				return;
			
			if(typeof sWord != "string")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_SEARCH_WORD, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setSearchWord");
			
			this.searchWord = sWord;
			this.searchIgnoreCase = bIgnoreCase;
			this.searchRegExp = bRegExp;
			
			this.input();
			
			if(this.searchWord.length)
			{
				this.textElement.selectionEnd = this.textElement.selectionStart;
				this.searchNext();
			}
		}
		
		//==================================================
		//	Search previous
		//==================================================
		searchPrev()
		{
			if((!this.searchWord) || (!this.searchWord.length))
				return;
			
			let iStart = 0;
			let iEnd = this.textElement.selectionStart;
			let oRegEx;
			let iIndex;
			let iMax = -1;
			let sSearchWord;
			let sMatched;
			
			if(this.searchRegExp)
				sSearchWord = this.searchWord;
			else
				sSearchWord = jaswin.escapeRegExp(this.searchWord);
			
			if(this.searchIgnoreCase)
				oRegEx = new RegExp(sSearchWord, "i");
			else
				oRegEx = new RegExp(sSearchWord);
			
			if(oRegEx)
			{
				while(true)
				{
					iIndex = this.textElement.value.substring(iStart, iEnd).search(oRegEx);
					if(iIndex == -1)
						break;
					
					saMatched = this.textElement.value.substring(iStart, iEnd).match(oRegEx);
					iMax = iStart + iIndex;
					iStart = iMax + 1;
				}
				
				if(iMax != -1)
				{
					this.textElement.selectionStart = iMax;
					this.textElement.selectionEnd = iMax
												+ saMatched[saMatched.length - 1].length;
					
					if((jaswin.BROWSER == "chrome")
					|| (jaswin.BROWSER == "safari")
					|| (jaswin.BROWSER == "opera"))
					{
						let iCnt;
						let iLineHeight = Math.ceil(parseInt(this.textElement.style.fontSize) * 4 / 3 * 1.1);
						let iHeight = this.textElement.clientHeight;
						if(!this.textElement.value.substring(0, this.textElement.selectionStart).match(/\n/g))
							iCnt = 1;
						else
							iCnt = this.textElement.value.substring(0, this.textElement.selectionStart).match(/\n/g).length + 1;
						
						if((iCnt + 1) * iLineHeight > iHeight)
						{
							this.textElement.scrollTop = (iCnt + 1) * iLineHeight - iHeight;
						}
					}
				}
			}
		}
		
		//==================================================
		//	Search next
		//==================================================
		searchNext()
		{
			if((!this.searchWord) || (!this.searchWord.length))
				return;
			
			let iStart = this.textElement.selectionEnd;
			let oRegEx;
			let iIndex;
			let sSearchWord;
			
			if(this.searchRegExp)
				sSearchWord = this.searchWord;
			else
				sSearchWord = jaswin.escapeRegExp(this.searchWord);
			
			if(this.searchIgnoreCase)
				oRegEx = new RegExp(sSearchWord, "i");
			else
				oRegEx = new RegExp(sSearchWord);
			
			if(oRegEx)
			{
				iIndex = this.textElement.value.substring(iStart).search(oRegEx);
				if(iIndex != -1)
				{
					this.textElement.selectionStart = iStart + iIndex;
					this.textElement.selectionEnd = iStart + iIndex
													+ this.textElement.value.substring(iStart).match(oRegEx)[0].length;
					
					if((jaswin.BROWSER == "chrome")
					|| (jaswin.BROWSER == "safari")
					|| (jaswin.BROWSER == "opera"))
					{
						let iCnt;
						let iLineHeight = Math.ceil(parseInt(this.textElement.style.fontSize) * 4 / 3 * 1.1);
						let iHeight = this.textElement.clientHeight;
						if(!this.textElement.value.substring(0, this.textElement.selectionStart).match(/\n/g))
							iCnt = 1;
						else
							iCnt = this.textElement.value.substring(0, this.textElement.selectionStart).match(/\n/g).length + 1;
						
						if((iCnt + 1) * iLineHeight > iHeight)
						{
							this.textElement.scrollTop = (iCnt + 1) * iLineHeight - iHeight;
						}
					}
				}
			}
		}
		
		//==================================================
		//	Set focus event
		//==================================================
		setFocusEvent(fpFocus)
		{
			if(typeof fpFocus != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_FOCUS, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setFocusEvent");
			this.onFocus = fpFocus;
		}
		
		//==================================================
		//	Set blur event
		//==================================================
		setBlurEvent(fpBlur)
		{
			if(typeof fpBlur != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_BLUR, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setBlurEvent");
			this.onBlur = fpBlur;
		}
		
		//==================================================
		//	Set keydown event
		//==================================================
		setKeyDownEvent(fpKeyDown)
		{
			if(typeof fpKeyDown != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_KEYDOWN, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setKeyDownEvent");
			this.onKeyDown = fpKeyDown;
		}
		
		//==================================================
		//	Set change event
		//==================================================
		setChangeEvent(fpChange)
		{
			if(typeof fpChange != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_CHANGE, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setChangeEvent");
			
			this.onChange = fpChange;
		}
		
		//==================================================
		//	Set input event
		//==================================================
		setInputEvent(fpInput)
		{
			if(typeof fpInput != "function")
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_ON_INPUT, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setInputEvent");
			this.onInput = fpInput;
		}
		
		//==================================================
		//	Set highlighting function
		//==================================================
		setHighlighting(fpHighlighting)
		{
			if(fpHighlighting)
			{
				if(typeof fpHighlighting != "function")
					throw new jaswin.JaswinError(jaswin.ERR_ALERT,
												jaswin.formatMessage(jaswin.ALERT_TYPE_ON_HIGHLIGHT, this.id),
												"textbox.js", "jaswin.textbox.EditBox.setHighlighting");
			}
			this.onHighlight = fpHighlighting;
			this.input();
		}
		
		//==================================================
		//	Set suggestion list
		//==================================================
		setSuggestionList(oListBox)
		{
			if(!jaswin.list)
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_INCLUDE, "list.js"),
											"textbox.js", "jaswin.textbox.EditBox.setSuggestionList");
			
			if((typeof oListBox != "object")
			|| (!oListBox instanceof jaswin.list.ListBox))
				throw new jaswin.JaswinError(jaswin.ERR_ALERT,
											jaswin.formatMessage(jaswin.ALERT_TYPE_LISTBOX, this.id),
											"textbox.js", "jaswin.textbox.EditBox.setSuggestionList");
			
			this.suggestionList = oListBox;
		}
	}
	
}(jaswin.textbox = jaswin.textbox || {}));
